@@ -6,6 +6,7 @@
 #include <osmocom/core/write_queue.h>
 #include <osmocom/core/timer.h>
 #include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/sockaddr_str.h>
 #include <osmocom/gsm/gsm23003.h>
 #include <osmocom/sigtran/sccp_sap.h>
 #include <osmocom/sigtran/osmo_ss7.h>
@@ -18,6 +19,8 @@
 #include <osmocom/mgcp_client/mgcp_client.h>
 #include <osmocom/mgcp_client/mgcp_client_pool.h>
 
+#include <osmocom/hnbgw/nft_kpi.h>
+
 #define STORE_UPTIME_INTERVAL	10 /* seconds */
 #define HNB_STORE_RAB_DURATIONS_INTERVAL 1 /* seconds */
 
@@ -29,6 +32,7 @@ enum {
 	DMGW,
 	DHNB,
 	DCN,
+	DNFT,
 };
 
 extern const struct log_info hnbgw_log_info;
@@ -37,6 +41,12 @@ extern struct vty_app_info hnbgw_vty_info;
 #define LOGHNB(HNB_CTX, ss, lvl, fmt, args ...) \
 	LOGP(ss, lvl, "(%s) " fmt, hnb_context_name(HNB_CTX), ## args)
 
+#define LOG_HNBP(HNBP, lvl, fmt, args...) \
+	LOGP(DHNB, lvl, "(%s) " fmt, \
+	     (HNBP) ? \
+		     (((HNBP)->id_str && *(HNBP)->id_str) ? (HNBP)->id_str : "no-cell-id") \
+		     : "null", ## args)
+
 #define DOMAIN_CS RANAP_CN_DomainIndicator_cs_domain
 #define DOMAIN_PS RANAP_CN_DomainIndicator_ps_domain
 
@@ -100,6 +110,8 @@ enum hnb_rate_ctr {
 
 	HNB_CTR_RANAP_PS_RAB_REL_REQ,
 	HNB_CTR_RANAP_CS_RAB_REL_REQ,
+	HNB_CTR_RANAP_PS_RAB_REL_REQ_ABNORMAL,
+	HNB_CTR_RANAP_CS_RAB_REL_REQ_ABNORMAL,
 
 	HNB_CTR_RANAP_PS_RAB_REL_CNF,
 	HNB_CTR_RANAP_CS_RAB_REL_CNF,
@@ -109,6 +121,8 @@ enum hnb_rate_ctr {
 
 	HNB_CTR_RANAP_PS_RAB_REL_IMPLICIT,
 	HNB_CTR_RANAP_CS_RAB_REL_IMPLICIT,
+	HNB_CTR_RANAP_PS_RAB_REL_IMPLICIT_ABNORMAL,
+	HNB_CTR_RANAP_CS_RAB_REL_IMPLICIT_ABNORMAL,
 
 	HNB_CTR_RUA_ERR_IND,
 
@@ -132,6 +146,25 @@ enum hnb_rate_ctr {
 	HNB_CTR_CS_PAGING_ATTEMPTED,
 
 	HNB_CTR_RAB_ACTIVE_MILLISECONDS_TOTAL,
+
+	HNB_CTR_DTAP_CS_LU_REQ,
+	HNB_CTR_DTAP_CS_LU_ACC,
+	HNB_CTR_DTAP_CS_LU_REJ,
+
+	HNB_CTR_DTAP_PS_ATT_REQ,
+	HNB_CTR_DTAP_PS_ATT_ACK,
+	HNB_CTR_DTAP_PS_ATT_REJ,
+
+	HNB_CTR_DTAP_PS_RAU_REQ,
+	HNB_CTR_DTAP_PS_RAU_ACK,
+	HNB_CTR_DTAP_PS_RAU_REJ,
+
+	HNB_CTR_GTPU_PACKETS_UL,
+	HNB_CTR_GTPU_TOTAL_BYTES_UL,
+	HNB_CTR_GTPU_UE_BYTES_UL,
+	HNB_CTR_GTPU_PACKETS_DL,
+	HNB_CTR_GTPU_TOTAL_BYTES_DL,
+	HNB_CTR_GTPU_UE_BYTES_DL,
 };
 
 enum hnb_stat {
@@ -139,22 +172,22 @@ enum hnb_stat {
 };
 
 struct umts_cell_id {
-	uint16_t mcc;	/*!< Mobile Country Code (0-999) */
-	uint16_t mnc;	/*!< Mobile Network Code (0-999) */
+	struct osmo_plmn_id plmn;	/*!< Mobile Country Code and Mobile Network Code (000-00 to 999-999) */
 	uint16_t lac;	/*!< Locaton Area Code (1-65534) */
 	uint16_t rac;	/*!< Routing Area Code (0-255) */
 	uint16_t sac;	/*!< Service Area Code */
 	uint32_t cid;	/*!< Cell ID */
 };
-const char *umts_cell_id_name(const struct umts_cell_id *ucid);
+int umts_cell_id_to_str_buf(char *buf, size_t buflen, const struct umts_cell_id *ucid);
+char *umts_cell_id_to_str_c(void *ctx, const struct umts_cell_id *ucid);
+const char *umts_cell_id_to_str(const struct umts_cell_id *ucid);
 int umts_cell_id_from_str(struct umts_cell_id *ucid, const char *instr);
+uint32_t umts_cell_id_hash(const struct umts_cell_id *ucid);
 
 /*! are both given umts_cell_id euqal? */
 static inline bool umts_cell_id_equal(const struct umts_cell_id *a, const struct umts_cell_id *b)
 {
-	if (a->mcc != b->mcc)
-		return false;
-	if (a->mnc != b->mnc)
+	if (osmo_plmn_cmp(&a->plmn, &b->plmn))
 		return false;
 	if (a->lac != b->lac)
 		return false;
@@ -340,6 +373,8 @@ struct hnb_context {
 struct hnb_persistent {
 	/*! Entry in HNBGW-global list of hnb_persistent */
 	struct llist_head list;
+	/*! Entry in hash table g_hnbgw->hnb_persistent_by_id. */
+	struct hlist_node node_by_id;
 	/*! back-pointer to hnb_context.  Can be NULL if no context at this point */
 	struct hnb_context *ctx;
 
@@ -353,6 +388,32 @@ struct hnb_persistent {
 
 	struct rate_ctr_group *ctrs;
 	struct osmo_stat_item_group *statg;
+
+	/* State that the main thread needs in order to know what was requested from the nft worker threads and what
+	 * still needs to be requested. */
+	struct {
+		/* Whether a persistent named counter was added in nftables for this cell id. */
+		bool persistent_counter_added;
+
+		/* The last hNodeB GTP-U address we asked the nft maintenance thread to set up.
+		 * osmo_sockaddr_str_is_nonzero(addr_remote) == false when no rules were added yet, and when
+		 * we asked the nft maintenance thread to remove the rules for this hNodeB because it has
+		 * disconnected. */
+		struct osmo_sockaddr_str addr_remote;
+
+		/* the nft handles needed to clean up the UL and DL rules when the hNodeB disconnects,
+		 * and the last seen counter value gotten from nft. */
+		struct {
+			struct nft_kpi_handle h;
+			struct nft_kpi_val v;
+		} ul;
+		struct {
+			struct nft_kpi_handle h;
+			struct nft_kpi_val v;
+		} dl;
+	} nft_kpi;
+
+	struct osmo_timer_list disconnected_timeout;
 };
 
 struct ue_context {
@@ -391,16 +452,24 @@ struct hnbgw {
 				char *core;
 			} netinst;
 		} pfcp;
+		struct {
+			bool enable;
+			/* The table name as used in nftables for the ruleset owned by this process. It is "osmo-hnbgw"
+			 * by default. */
+			char *table_name;
+		} nft_kpi;
 	} config;
 	/*! SCTP listen socket for incoming connections */
 	struct osmo_stream_srv_link *iuh;
 	/* list of struct hnb_context */
 	struct llist_head hnb_list;
+
 	/* list of struct hnb_persistent */
 	struct llist_head hnb_persistent_list;
+	/* optimized lookup for hnb_persistent, by cell id string */
+	DECLARE_HASHTABLE(hnb_persistent_by_id, 5);
+
 	struct osmo_timer_list store_uptime_timer;
-	/* list of struct ue_context */
-	struct llist_head ue_list;
 	/* next availble UE Context ID */
 	uint32_t next_ue_ctx_id;
 	struct ctrl_handle *ctrl;
@@ -424,6 +493,12 @@ struct hnbgw {
 	} pfcp;
 
 	struct osmo_timer_list hnb_store_rab_durations_timer;
+
+	struct {
+		bool active;
+		struct osmo_timer_list get_counters_timer;
+		struct timespec next_timer;
+	} nft_kpi;
 };
 
 extern struct hnbgw *g_hnbgw;
@@ -439,18 +514,13 @@ int hnbgw_mgw_setup(void);
 struct hnb_context *hnb_context_by_identity_info(const char *identity_info);
 const char *hnb_context_name(struct hnb_context *ctx);
 
-struct ue_context *ue_context_by_id(uint32_t id);
-struct ue_context *ue_context_by_imsi(const char *imsi);
-struct ue_context *ue_context_by_tmsi(uint32_t tmsi);
-struct ue_context *ue_context_alloc(struct hnb_context *hnb, const char *imsi,
-				    uint32_t tmsi);
-void ue_context_free(struct ue_context *ue);
-
 void hnb_context_release(struct hnb_context *ctx);
 void hnb_context_release_ue_state(struct hnb_context *ctx);
 
 struct hnb_persistent *hnb_persistent_alloc(const struct umts_cell_id *id);
 struct hnb_persistent *hnb_persistent_find_by_id(const struct umts_cell_id *id);
+void hnb_persistent_registered(struct hnb_persistent *hnbp);
+void hnb_persistent_deregistered(struct hnb_persistent *hnbp);
 void hnb_persistent_free(struct hnb_persistent *hnbp);
 
 void hnbgw_vty_init(void);
@@ -471,3 +541,5 @@ struct msgb *hnbgw_ranap_msg_alloc(const char *name);
 int hnbgw_peek_l3_ul(struct hnbgw_context_map *map, struct msgb *ranap_msg);
 
 unsigned long long hnb_get_updowntime(const struct hnb_context *ctx);
+
+uint32_t get_next_ue_ctx_id(void);
