@@ -22,6 +22,7 @@
 #include <osmocom/core/utils.h>
 #include <osmocom/core/socket.h>
 #include <osmocom/gsm/gsm48.h>
+#include <osmocom/gsm/protocol/gsm_23_003.h>
 #include <osmocom/netif/stream.h>
 
 #include <unistd.h>
@@ -170,7 +171,7 @@ static int hnbgw_tx_hnb_register_acc(struct hnb_context *ctx)
 }
 
 
-static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
+static int hnbgw_tx_ue_register_acc(struct hnb_context *hnb, const char *imsi, uint32_t context_id)
 {
 	HNBAP_UERegisterAccept_t accept_out;
 	HNBAP_UERegisterAcceptIEs_t accept;
@@ -181,18 +182,20 @@ static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
 	int rc;
 
 	encoded_imsi_len = ranap_imsi_encode(encoded_imsi,
-					  sizeof(encoded_imsi), ue->imsi);
+					  sizeof(encoded_imsi), imsi);
 
 	memset(&accept, 0, sizeof(accept));
 	accept.uE_Identity.present = HNBAP_UE_Identity_PR_iMSI;
 	OCTET_STRING_fromBuf(&accept.uE_Identity.choice.iMSI,
 			     (const char *)encoded_imsi, encoded_imsi_len);
-	asn1_u24_to_bitstring(&accept.context_ID, &ctx_id, ue->context_id);
+	asn1_u24_to_bitstring(&accept.context_ID, &ctx_id, context_id);
 
 	memset(&accept_out, 0, sizeof(accept_out));
 	rc = hnbap_encode_ueregisteraccepties(&accept_out, &accept);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_OCTET_STRING, &accept.uE_Identity.choice.iMSI);
 	if (rc < 0) {
+		LOGHNB(hnb, DHNBAP, LOGL_ERROR,
+		       "Failed to encode HNBAP UE Register Accept message for UE IMSI-%s\n", imsi);
 		return rc;
 	}
 
@@ -203,13 +206,18 @@ static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
 
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_UERegisterAccept, &accept_out);
 
-	return hnbgw_hnbap_tx(ue->hnb, msg);
+	rc = hnbgw_hnbap_tx(hnb, msg);
+	if (rc)
+		LOGHNB(hnb, DHNBAP, LOGL_ERROR,
+		       "Failed to enqueue HNBAP UE Register Accept message for UE IMSI-%s\n", imsi);
+	return rc;
 }
 
 static int hnbgw_tx_ue_register_rej(struct hnb_context *hnb, HNBAP_UE_Identity_t *ue_id, const struct HNBAP_Cause *cause)
 {
 	HNBAP_UERegisterReject_t reject_out;
 	HNBAP_UERegisterRejectIEs_t reject;
+	char imsi[GSM23003_IMSI_MAX_DIGITS+1];
 	struct msgb *msg;
 	int rc;
 
@@ -270,6 +278,14 @@ static int hnbgw_tx_ue_register_rej(struct hnb_context *hnb, HNBAP_UE_Identity_t
 				     ue_id->choice.pTMSIRAI.rAI.rAC.size);
 		break;
 
+	case HNBAP_UE_Identity_PR_iMSI:
+		ranap_bcd_decode(imsi, sizeof(imsi), ue_id->choice.iMSI.buf, ue_id->choice.iMSI.size);
+		LOGHNB(hnb, DHNBAP, LOGL_DEBUG, "REJ UE_Id IMSI %s\n", imsi);
+
+		OCTET_STRING_fromBuf(&reject.uE_Identity.choice.iMSI,
+				     (const char *)ue_id->choice.iMSI.buf, ue_id->choice.iMSI.size);
+		break;
+
 	default:
 		LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Cannot compose UE Register Reject:"
 		     " unsupported UE ID (present=%d)\n", ue_id->present);
@@ -311,6 +327,9 @@ static int hnbgw_tx_ue_register_rej(struct hnb_context *hnb, HNBAP_UE_Identity_t
 		ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_OCTET_STRING,
 					      &reject.uE_Identity.choice.pTMSIRAI.rAI.rAC);
 		break;
+	case HNBAP_UE_Identity_PR_iMSI:
+		ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_OCTET_STRING,
+					      &reject.uE_Identity.choice.iMSI);
 
 	default:
 		/* should never happen after above switch() */
@@ -329,8 +348,6 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 	struct msgb *msg;
 	uint32_t ctx_id;
 	uint32_t tmsi = 0;
-	struct ue_context *ue;
-	struct ue_context *ue_allocated = NULL;
 	int rc;
 
 	memset(&accept, 0, sizeof(accept));
@@ -376,11 +393,7 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 	tmsi = ntohl(tmsi);
 	LOGHNB(hnb, DHNBAP, LOGL_DEBUG, "HNBAP register with TMSI %x\n", tmsi);
 
-	ue = ue_context_by_tmsi(tmsi);
-	if (!ue)
-		ue = ue_allocated = ue_context_alloc(hnb, NULL, tmsi);
-
-	asn1_u24_to_bitstring(&accept.context_ID, &ctx_id, ue->context_id);
+	asn1_u24_to_bitstring(&accept.context_ID, &ctx_id, get_next_ue_ctx_id());
 
 	memset(&accept_out, 0, sizeof(accept_out));
 	rc = hnbap_encode_ueregisteraccepties(&accept_out, &accept);
@@ -413,11 +426,8 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 	}
 
 	if (rc < 0) {
+		LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Failed to encode HNBAP UE Register Accept for TMSI 0x%08x\n", tmsi);
 		/* Encoding failed. Nothing in 'accept_out'. */
-		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
-		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
-		if (ue_allocated)
-			ue_context_free(ue_allocated);
 		return rc;
 	}
 
@@ -428,6 +438,8 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 						&accept_out);
 	rc = hnbgw_hnbap_tx(hnb, msg);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_UERegisterAccept, &accept_out);
+	if (rc)
+		LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Failed to transmit HNBAP UE Register Accept for TMSI 0x%08x\n", tmsi);
 	return rc;
 }
 
@@ -459,12 +471,13 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	struct hnb_context *hnb, *tmp;
 	HNBAP_HNBRegisterRequestIEs_t ies;
 	int rc;
-	struct osmo_plmn_id plmn;
 	struct osmo_fd *ofd = osmo_stream_srv_get_ofd(ctx->conn);
 	char identity_str[256];
 	const char *cell_id_str;
 	struct timespec tp;
 	HNBAP_Cause_t cause;
+	struct osmo_sockaddr cur_osa = {};
+	socklen_t len = sizeof(cur_osa);
 
 	rc = hnbap_decode_hnbregisterrequesties(&ies, in);
 	if (rc < 0) {
@@ -477,14 +490,25 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 
 	/* copy all identity parameters from the message to ctx */
 	OSMO_STRLCPY_ARRAY(ctx->identity_info, identity_str);
+
+	/* We want to use struct umts_cell_id as hashtable key. If it ever happens to contain any padding bytes, make
+	 * sure everything is deterministically zero. */
+	memset(&ctx->id, 0, sizeof(ctx->id));
 	ctx->id.lac = asn1str_to_u16(&ies.lac);
 	ctx->id.sac = asn1str_to_u16(&ies.sac);
 	ctx->id.rac = asn1str_to_u8(&ies.rac);
 	ctx->id.cid = asn1bitstr_to_u28(&ies.cellIdentity);
-	osmo_plmn_from_bcd(ies.plmNidentity.buf, &plmn);
-	ctx->id.mcc = plmn.mcc;
-	ctx->id.mnc = plmn.mnc;
-	cell_id_str = umts_cell_id_name(&ctx->id);
+	osmo_plmn_from_bcd(ies.plmNidentity.buf, &ctx->id.plmn);
+	cell_id_str = umts_cell_id_to_str(&ctx->id);
+
+	if (getpeername(ofd->fd, &cur_osa.u.sa, &len) < 0) {
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "HNB-REGISTER-REQ %s: rejecting due to getpeername() error: %s\n",
+		       cell_id_str, strerror(errno));
+		hnbap_free_hnbregisterrequesties(&ies);
+		cause.present = HNBAP_Cause_PR_radioNetwork;
+		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_hNB_parameter_mismatch;
+		return hnbgw_tx_hnb_register_rej(ctx, &cause);
+	}
 
 	hnbp = hnb_persistent_find_by_id(&ctx->id);
 	if (!hnbp && g_hnbgw->config.accept_all_hnb)
@@ -497,6 +521,7 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_unauthorised_HNB;
 		return hnbgw_tx_hnb_register_rej(ctx, &cause);
 	}
+
 	ctx->persistent = hnbp;
 	hnbp->ctx = ctx;
 
@@ -510,22 +535,13 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 			 * fault (bug), and we release the old context to keep going... */
 			struct osmo_fd *other_fd = osmo_stream_srv_get_ofd(hnb->conn);
 			struct osmo_sockaddr other_osa = {};
-			struct osmo_sockaddr cur_osa = {};
 			socklen_t len = sizeof(other_osa);
 			if (getpeername(other_fd->fd, &other_osa.u.sa, &len) < 0) {
 				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with invalid socket, releasing it\n");
 				hnb_context_release(hnb);
 				continue;
 			}
-			len = sizeof(cur_osa);
-			if (getpeername(ofd->fd, &cur_osa.u.sa, &len) < 0) {
-				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Error getpeername(): %s\n", strerror(errno));
-				if (osmo_sockaddr_cmp(&cur_osa, &other_osa) == 0) {
-					LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with same remote address, releasing it\n");
-					hnb_context_release(hnb);
-					continue;
-				}
-			} else if (osmo_sockaddr_cmp(&cur_osa, &other_osa) == 0) {
+			if (osmo_sockaddr_cmp(&cur_osa, &other_osa) == 0) {
 				LOGHNB(ctx, DHNBAP, LOGL_ERROR, "BUG! Found old registered HNB with same remote address, releasing it\n");
 				hnb_context_release(hnb);
 				continue;
@@ -552,6 +568,8 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 
 	ctx->hnb_registered = true;
 
+	hnb_persistent_registered(ctx->persistent);
+
 	/* Send HNBRegisterAccept */
 	rc = hnbgw_tx_hnb_register_acc(ctx);
 	hnbap_free_hnbregisterrequesties(&ies);
@@ -562,9 +580,7 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_UERegisterRequestIEs_t ies;
 	HNBAP_Cause_t cause;
-	struct ue_context *ue;
-	struct ue_context *ue_allocated = NULL;
-	char imsi[16];
+	char imsi[GSM23003_IMSI_MAX_DIGITS+1];
 	int rc;
 
 	rc = hnbap_decode_ueregisterrequesties(&ies, in);
@@ -620,18 +636,10 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-REGISTER-REQ ID_type=%d imsi=%s cause=%ld\n",
 		ies.uE_Identity.present, imsi, ies.registration_Cause);
 
-	ue = ue_context_by_imsi(imsi);
-	if (!ue)
-		ue = ue_allocated = ue_context_alloc(ctx, imsi, 0);
-
 	/* Send UERegisterAccept */
-	rc = hnbgw_tx_ue_register_acc(ue);
-	if (rc < 0) {
-		/* If we allocated the UE context but the UE REGISTER fails, get rid of it again: there will likely
-		 * never be a UE DE-REGISTER for this UE from the HNB, and the ue_context would linger forever. */
-		if (ue_allocated)
-			ue_context_free(ue_allocated);
-	}
+	rc = hnbgw_tx_ue_register_acc(ctx, imsi, get_next_ue_ctx_id());
+	if (rc < 0)
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failed to transmit HNBAP UE Register Accept for IMSI %s\n", imsi);
 free_and_return_rc:
 	hnbap_free_ueregisterrequesties(&ies);
 	return rc;
@@ -641,7 +649,6 @@ static int hnbgw_rx_ue_deregister(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_UEDe_RegisterIEs_t ies;
 	HNBAP_Cause_t cause;
-	struct ue_context *ue;
 	int rc;
 	uint32_t ctxid;
 
@@ -664,9 +671,6 @@ static int hnbgw_rx_ue_deregister(struct hnb_context *ctx, ANY_t *in)
 				   HNBAP_Criticality_ignore, HNBAP_TriggeringMessage_initiating_message);
 	} else {
 		LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-DE-REGISTER context=%u cause=%s\n", ctxid, hnbap_cause_str(&ies.cause));
-		ue = ue_context_by_id(ctxid);
-		if (ue)
-			ue_context_free(ue);
 	}
 
 	hnbap_free_uede_registeries(&ies);
@@ -739,8 +743,7 @@ static int hnbgw_rx_unsuccessful_outcome_msg(struct hnb_context *hnb, HNBAP_Unsu
 {
 	/* We don't care much about HNBAP */
 	LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Received Unsuccessful Outcome, procedureCode %ld, criticality %ld,"
-		" cell mcc %u mnc %u lac %u rac %u sac %u cid %u\n", msg->procedureCode, msg->criticality,
-		hnb->id.mcc, hnb->id.mnc, hnb->id.lac, hnb->id.rac, hnb->id.sac, hnb->id.cid);
+		" cell %s\n", msg->procedureCode, msg->criticality, umts_cell_id_to_str(&hnb->id));
 	return 0;
 }
 
