@@ -83,6 +83,7 @@ int rua_tx_udt(struct hnb_context *hnb, const uint8_t *data, unsigned int len)
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RUA_ConnectionlessTransfer, &out);
 
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA payload of %u bytes\n", msgb_length(msg));
+	HNBP_CTR_INC(hnb->persistent, HNB_CTR_RUA_UDT_DL);
 
 	return hnbgw_rua_tx(hnb, msg);
 }
@@ -120,6 +121,7 @@ int rua_tx_dt(struct hnb_context *hnb, int is_ps, uint32_t context_id,
 
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA DirectTransfer (cn=%s) payload of %u bytes\n",
 		is_ps ? "ps" : "cs", msgb_length(msg));
+	HNBP_CTR_INC(hnb->persistent, is_ps ? HNB_CTR_RUA_PS_DT_DL : HNB_CTR_RUA_CS_DT_DL);
 
 	return hnbgw_rua_tx(hnb, msg);
 }
@@ -157,9 +159,9 @@ int rua_tx_disc(struct hnb_context *hnb, int is_ps, uint32_t context_id,
 					      &out);
 	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RUA_Disconnect, &out);
 
-
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "transmitting RUA Disconnect (cn=%s) payload of %u bytes\n",
 	       is_ps ? "ps" : "cs", msgb_length(msg));
+	HNBP_CTR_INC(hnb->persistent, is_ps ? HNB_CTR_RUA_PS_DISCONNECT_DL : HNB_CTR_RUA_CS_DISCONNECT_DL);
 
 	return hnbgw_rua_tx(hnb, msg);
 }
@@ -179,18 +181,12 @@ static inline const char *rua_procedure_code_name(enum RUA_ProcedureCode val)
 	return get_value_string(rua_procedure_code_names, val);
 }
 
-static struct hnbgw_context_map *find_or_create_context_map(struct hnb_context *hnb, uint32_t rua_ctx_id, bool is_ps,
-							    struct msgb *ranap_msg)
+static struct hnbgw_context_map *create_context_map(struct hnb_context *hnb, uint32_t rua_ctx_id, bool is_ps,
+						    struct msgb *ranap_msg)
 {
 	struct hnbgw_context_map *map;
 	struct hnbgw_cnlink *cnlink;
 
-	map = context_map_find_by_rua_ctx_id(hnb, rua_ctx_id, is_ps);
-	if (map) {
-		/* Already established SCCP link. Continue to use that. */
-		return map;
-	}
-
 	/* Establish a new context map. From the RUA Connect, extract a mobile identity, if any, and select a CN link
 	 * based on an NRI found in the mobile identity, if any. */
 
@@ -198,7 +194,7 @@ static struct hnbgw_context_map *find_or_create_context_map(struct hnb_context *
 	map = context_map_alloc(hnb, rua_ctx_id, is_ps);
 	OSMO_ASSERT(map);
 
-	if (hnbgw_peek_l3(map, ranap_msg))
+	if (hnbgw_peek_l3_ul(map, ranap_msg))
 		LOGP(DCN, LOGL_NOTICE, "Failed to extract Mobile Identity from RUA Connect message's RANAP payload\n");
 	/* map->l3 now contains all the interesting information from the NAS PDU, if any.
 	 * If no useful information could be decoded, still continue to select a hopefully adequate link by round robin.
@@ -222,7 +218,7 @@ static struct hnbgw_context_map *find_or_create_context_map(struct hnb_context *
 
 /* dispatch a RUA connection-oriented message received from a HNB to a context mapping's RUA FSM, so that it is
  * forwarded to the CN via SCCP connection-oriented messages.
- * Connectionless messages are handled in hnbgw_ranap_rx() instead, not here. */
+ * Connectionless messages are handled in hnbgw_ranap_rx_udt_ul() instead, not here. */
 static int rua_to_scu(struct hnb_context *hnb,
 		      RUA_CN_DomainIndicator_t cN_DomainIndicator,
 		      enum RUA_ProcedureCode rua_procedure,
@@ -232,6 +228,7 @@ static int rua_to_scu(struct hnb_context *hnb,
 	struct msgb *ranap_msg = NULL;
 	struct hnbgw_context_map *map = NULL;
 	bool is_ps;
+	int logl;
 
 	switch (cN_DomainIndicator) {
 	case RUA_CN_DomainIndicator_cs_domain:
@@ -255,12 +252,42 @@ static int rua_to_scu(struct hnb_context *hnb,
 		memcpy(ranap_msg->l2h, data, len);
 	}
 
-	map = find_or_create_context_map(hnb, context_id, is_ps, ranap_msg);
-	if (!map) {
-		LOGHNB(hnb, DRUA, LOGL_ERROR,
-		       "Failed to create context map for %s: rx RUA %s with %u bytes RANAP data\n",
-		       is_ps ? "IuPS" : "IuCS", rua_procedure_code_name(rua_procedure), data ? len : 0);
-		return -1;
+	map = context_map_find_by_rua_ctx_id(hnb, context_id, is_ps);
+	logl = LOGL_ERROR;
+
+	switch (rua_procedure) {
+	case RUA_ProcedureCode_id_Connect:
+		/* A Connect message can only be the first message for an unused RUA context */
+		if (map) {
+			/* Already established this RUA context. But then how can it be a Connect message. */
+			LOGHNB(hnb, DRUA, LOGL_ERROR, "rx RUA %s for already active RUA context %u\n",
+			       rua_procedure_code_name(rua_procedure), context_id);
+			return -EINVAL;
+		}
+		/* ok, this RUA context does not exist yet, so create one. */
+		map = create_context_map(hnb, context_id, is_ps, ranap_msg);
+		if (!map) {
+			LOGHNB(hnb, DRUA, LOGL_ERROR,
+			       "Failed to create context map for %s: rx RUA %s with %u bytes RANAP data\n",
+			       is_ps ? "IuPS" : "IuCS", rua_procedure_code_name(rua_procedure), data ? len : 0);
+			return -EINVAL;
+		}
+		break;
+
+	case RUA_ProcedureCode_id_Disconnect:
+		/* For RUA Disconnect, do not spam the ERROR log. It is just a stray Disconnect, no harm done.
+		 * Context: some CN are known to rapidly tear down SCCP without waiting for RUA to disconnect gracefully
+		 * (IU Release Complete). Such CN would cause ERROR logging for each and every released context map. */
+		logl = LOGL_DEBUG;
+		/* fall thru */
+	default:
+		/* Any message other than Connect must have a valid RUA context */
+		if (!map) {
+			LOGHNB(hnb, DRUA, logl, "rx RUA %s for unknown RUA context %u\n",
+			       rua_procedure_code_name(rua_procedure), context_id);
+			return -EINVAL;
+		}
+		break;
 	}
 
 	LOG_MAP(map, DRUA, LOGL_DEBUG, "rx RUA %s with %u bytes RANAP data\n",
@@ -361,6 +388,8 @@ static int rua_rx_init_connect(struct msgb *msg, ANY_t *in)
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA %s Connect.req(ctx=0x%x, %s)\n",
 		ranap_domain_name(ies.cN_DomainIndicator), context_id,
 		ies.establishment_Cause == RUA_Establishment_Cause_emergency_call ? "emergency" : "normal");
+	HNBP_CTR_INC(hnb->persistent, ies.cN_DomainIndicator == DOMAIN_PS ?
+		     HNB_CTR_RUA_PS_CONNECT_UL : HNB_CTR_RUA_CS_CONNECT_UL);
 
 	rc = rua_to_scu(hnb, ies.cN_DomainIndicator, RUA_ProcedureCode_id_Connect,
 			context_id, 0, ies.ranaP_Message.buf,
@@ -390,6 +419,8 @@ static int rua_rx_init_disconnect(struct msgb *msg, ANY_t *in)
 
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA Disconnect.req(ctx=0x%x,cause=%s)\n", context_id,
 		rua_cause_str(&ies.cause));
+	HNBP_CTR_INC(hnb->persistent, ies.cN_DomainIndicator == DOMAIN_PS ?
+		     HNB_CTR_RUA_PS_DISCONNECT_UL : HNB_CTR_RUA_CS_DISCONNECT_UL);
 
 	if (ies.presenceMask & DISCONNECTIES_RUA_RANAP_MESSAGE_PRESENT) {
 		ranap_data = ies.ranaP_Message.buf;
@@ -419,6 +450,8 @@ static int rua_rx_init_dt(struct msgb *msg, ANY_t *in)
 	context_id = asn1bitstr_to_u24(&ies.context_ID);
 
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA Data.req(ctx=0x%x)\n", context_id);
+	HNBP_CTR_INC(hnb->persistent, ies.cN_DomainIndicator == DOMAIN_PS ?
+		     HNB_CTR_RUA_PS_DT_UL : HNB_CTR_RUA_CS_DT_UL);
 
 	rc = rua_to_scu(hnb,
 			ies.cN_DomainIndicator,
@@ -442,6 +475,7 @@ static int rua_rx_init_udt(struct msgb *msg, ANY_t *in)
 		return rc;
 
 	LOGHNB(hnb, DRUA, LOGL_DEBUG, "RUA UData.req()\n");
+	HNBP_CTR_INC(hnb->persistent, HNB_CTR_RUA_UDT_UL);
 
 	/* according tot the spec, we can primarily receive Overload,
 	 * Reset, Reset ACK, Error Indication, reset Resource, Reset
@@ -450,7 +484,7 @@ static int rua_rx_init_udt(struct msgb *msg, ANY_t *in)
 	 * Information Transfer and Uplink Information Trnansfer that we
 	 * can ignore.  In either case, it is RANAP that we need to
 	 * decode... */
-	rc = hnbgw_ranap_rx(msg, ies.ranaP_Message.buf, ies.ranaP_Message.size);
+	rc = hnbgw_ranap_rx_udt_ul(msg, ies.ranaP_Message.buf, ies.ranaP_Message.size);
 	rua_free_connectionlesstransferies(&ies);
 
 	return rc;
@@ -468,6 +502,7 @@ static int rua_rx_init_err_ind(struct msgb *msg, ANY_t *in)
 		return rc;
 
 	LOGHNB(hnb, DRUA, LOGL_ERROR, "RUA UData.ErrorInd(%s)\n", rua_cause_str(&ies.cause));
+	HNBP_CTR_INC(hnb->persistent, HNB_CTR_RUA_ERR_IND);
 
 	rua_free_errorindicationies(&ies);
 	return rc;
@@ -554,6 +589,10 @@ int hnbgw_rua_rx(struct hnb_context *hnb, struct msgb *msg)
 	asn_dec_rval_t dec_ret;
 	int rc;
 
+	/* RUA is only processed after HNB registration, and as soon as the HNB is registered,
+	 * it should have a persistent config associated with it */
+	OSMO_ASSERT(hnb->persistent);
+
 	/* decode and handle to _hnbgw_hnbap_rx() */
 
 	memset(pdu, 0, sizeof(*pdu));
