@@ -1,4 +1,4 @@
-/* (C) 2021 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
+/* (C) 2021-2024 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * Author: Philipp Maier
@@ -47,9 +47,9 @@
 #include <osmocom/hnbgw/tdefs.h>
 #include <osmocom/mgcp_client/mgcp_client_endpoint_fsm.h>
 
-/* NOTE: This implementation can only handle one RAB per hnbgw context. This simplification was made because usually
- * a voice call will require only one RAB at a time. An exception may be corner cases like video calls, which we
- * do not support at the moment. */
+/* NOTE: This implementation can only handle one RAB per context_map (== SCCP connection == RUA connection == UE
+ * context). This simplification was made because usually a voice call will require only one RAB at a time. An exception
+ * may be corner cases like video calls, which we do not support at the moment. */
 
 /* Send Iu Release Request, this is done in erroneous cases from which we cannot recover */
 static void tx_release_req(struct hnbgw_context_map *map)
@@ -123,6 +123,10 @@ struct mgw_fsm_priv {
 	struct osmo_sockaddr ci_hnb_crcx_ack_addr;
 	char msc_rtp_addr[INET6_ADDRSTRLEN];
 	uint16_t msc_rtp_port;
+
+	/* Timestamps to track active duration */
+	struct timespec active_start;
+	struct timespec active_stored;
 };
 
 struct osmo_tdef_state_timeout mgw_fsm_timeouts[32] = {
@@ -171,7 +175,7 @@ static void mgw_fsm_crcx_hnb_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	mgw_info = (struct mgcp_conn_peer) {
 		.call_id = (map->rua_ctx_id << 8) | mgw_fsm_priv->rab_id,
 		.ptime = 20,
-		.conn_mode = MGCP_CONN_LOOPBACK,
+		.conn_mode = MGCP_CONN_RECV_ONLY,
 	};
 	mgw_info.codecs[0] = CODEC_IUFP;
 	mgw_info.codecs_len = 1;
@@ -554,6 +558,9 @@ static void mgw_fsm_established_onenter(struct osmo_fsm_inst *fi, uint32_t prev_
 	}
 
 	LOGPFSML(fi, LOGL_DEBUG, "HNB and MSC side call-legs completed!\n");
+
+	osmo_clock_gettime(CLOCK_MONOTONIC, &mgw_fsm_priv->active_start);
+	mgw_fsm_priv->active_stored = mgw_fsm_priv->active_start;
 }
 
 static void mgw_fsm_release_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
@@ -794,12 +801,14 @@ static bool is_our_rab_release(struct hnbgw_context_map *map, ranap_message *mes
  *  \param[in] message decoded RANAP message container, allocated in OTC_SELECT.
  *                     This function may talloc_steal(message) to keep it for later.
  *  \returns 0 on success; negative on error. */
-int handle_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
+int handle_cs_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	static bool initialized = false;
 	struct mgw_fsm_priv *mgw_fsm_priv;
 	char fsm_name[255];
 
+	OSMO_ASSERT(!map->is_ps);
+
 	/* Initialize FSM if not done yet */
 	if (!initialized) {
 		OSMO_ASSERT(osmo_fsm_register(&mgw_fsm) == 0);
@@ -866,11 +875,12 @@ int handle_rab_ass_req(struct hnbgw_context_map *map, struct msgb *ranap_msg, ra
  *  \param[in] message decoded RANAP message container, allocated in OTC_SELECT.
  *                     This function may talloc_steal(message) to keep it for later.
  *  \returns 0 on success; negative on error. */
-int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
+int mgw_fsm_handle_cs_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *ranap_msg, ranap_message *message)
 {
 	struct mgw_fsm_priv *mgw_fsm_priv;
 
 	OSMO_ASSERT(ranap_msg);
+	OSMO_ASSERT(!map->is_ps);
 
 	if (!map->mgw_fi) {
 		/* NOTE: This situation is a corner-case. We may end up here when the co-located MGW caused a problem
@@ -909,9 +919,40 @@ int mgw_fsm_handle_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *rana
  *  \returns 0 on success; negative on error. */
 int mgw_fsm_release(struct hnbgw_context_map *map)
 {
+	OSMO_ASSERT(!map->is_ps);
+
 	if (!map->mgw_fi)
 		return -EINVAL;
 
 	osmo_fsm_inst_dispatch(map->mgw_fi, MGW_EV_RELEASE, NULL);
 	return 0;
 }
+
+/* determine the number of elapsed active RAB milli-seconds since last call */
+uint64_t mgw_fsm_get_elapsed_ms(struct hnbgw_context_map *map, const struct timespec *now)
+{
+	struct mgw_fsm_priv *mgw_fsm_priv;
+	struct timespec elapsed;
+	uint64_t elapsed_ms;
+
+	OSMO_ASSERT(!map->is_ps);
+
+	if (!map->mgw_fi)
+		return 0;
+
+	OSMO_ASSERT(map->mgw_fi->fsm == &mgw_fsm);
+	mgw_fsm_priv = map->mgw_fi->priv;
+
+	/* Ignore RABs whose activation timestamps are not yet set. */
+	if (!timespecisset(&mgw_fsm_priv->active_stored))
+		return 0;
+
+	/* Calculate elapsed time since last storage */
+	timespecsub(now, &mgw_fsm_priv->active_stored, &elapsed);
+	elapsed_ms = elapsed.tv_sec * 1000 + elapsed.tv_nsec / 1000000;
+
+	/* Update storage time */
+	mgw_fsm_priv->active_stored = *now;
+
+	return elapsed_ms;
+}
