@@ -1,6 +1,6 @@
 /* OsmoHNBGW main routine */
 
-/* (C) 2015 by Harald Welte <laforge@gnumonks.org>
+/* (C) 2015-2024 by Harald Welte <laforge@gnumonks.org>
  * (C) 2016-2023 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
@@ -19,6 +19,7 @@
  *
  */
 
+#include <signal.h>
 #include <getopt.h>
 
 #include "config.h"
@@ -64,6 +65,32 @@ static struct {
 	NULL,
 };
 
+static void signal_handler(int signum)
+{
+	fprintf(stdout, "signal %u received\n", signum);
+
+	switch (signum) {
+	case SIGABRT:
+		/* in case of abort, we want to obtain a talloc report and
+		 * then run default SIGABRT handler, who will generate coredump
+		 * and abort the process. abort() should do this for us after we
+		 * return, but program wouldn't exit if an external SIGABRT is
+		 * received.
+		 */
+		talloc_report(tall_vty_ctx, stderr);
+		talloc_report_full(g_hnbgw, stderr);
+		signal(SIGABRT, SIG_DFL);
+		raise(SIGABRT);
+		break;
+	case SIGUSR1:
+		talloc_report(tall_vty_ctx, stderr);
+		talloc_report_full(g_hnbgw, stderr);
+		break;
+	default:
+		break;
+	}
+}
+
 static void print_usage(void)
 {
 	printf("Usage: osmo-hnbgw\n");
@@ -195,6 +222,7 @@ int main(int argc, char **argv)
 	rc = osmo_init_logging2(g_hnbgw, &hnbgw_log_info);
 	if (rc < 0)
 		exit(1);
+	log_enable_multithread();
 
 	osmo_stats_init(g_hnbgw);
 	rc = rate_ctr_init(g_hnbgw);
@@ -302,6 +330,20 @@ int main(int argc, char **argv)
 	hnbgw_pfcp_init();
 #endif
 
+	/* If nftables is enabled, initialize the nft table now or fail startup. This is important to immediately let
+	 * the user know if cap_net_admin privileges are missing, and not only when the first hNodeB connects. */
+	if (g_hnbgw->config.nft_kpi.enable) {
+#if ENABLE_NFTABLES
+		nft_kpi_init(g_hnbgw->config.nft_kpi.table_name);
+		/* There is no direct error handling here, because nftables initialization happens asynchronously.
+		 * See nft_kpi.c nft_thread_t2m_cb(), case NFT_THREAD_INIT_TABLE to see what happens when initializing
+		 * nftables failed. */
+#else
+		fprintf(stderr, "ERROR: Cannot enable nft KPI, this binary was built without nftables support\n");
+		exit(1);
+#endif
+	}
+
 	hnbgw_cnpool_start(&g_hnbgw->sccp.cnpool_iucs);
 	hnbgw_cnpool_start(&g_hnbgw->sccp.cnpool_iups);
 
@@ -313,6 +355,13 @@ int main(int argc, char **argv)
 		}
 	}
 
+	signal(SIGABRT, &signal_handler);
+	signal(SIGUSR1, &signal_handler);
+	signal(SIGUSR2, &signal_handler);
+	osmo_init_ignore_signals();
+
+	osmo_fsm_set_dealloc_ctx(OTC_SELECT);
+
 	while (1) {
 		rc = osmo_select_main_ctx(0);
 		if (rc < 0)
